/* 
 * tigra.c - Computer Graphics, Autumn 2000
 *
 * Author: Ilpo Jrvinen <ilpo.jarvinen@helsinki.fi>
 *
 * Requirements: MesaGL, other opengl implementations might require tweaking.
 *
 * Description:
 *    This program creates candlesticks on the randomly textured table and
 *    displays sweeping convey-hull finder (a.k.a Graham's algorithm).
 *    The sweeper will sweep the are infinitely.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <sys/time.h>
#include <GL/glut.h>



/* Candle area */
#define AREA_MINX -1.8
#define AREA_MINZ -0.8
#define AREA_MAXX 1.8
#define AREA_MAXZ 0.8

/* candle definations */
#define POINTS 8
#define KERNEL_LEN 0.1
#define CANDLE_TOP 0.95
#define CANDLE_UPPER 1.0
#define CANDLE_LOWER 0.0
#define CANDLE_RADIUS 0.03
#define KERNEL_RADIUS (CANDLE_RADIUS/6)
#define CANDLE_QUADS 16
#define SWEEP_RADIUS 0.032

/* sweep parameters */
#define SWEEP_LOWER 0.2
#define SWEEP_UPPER 0.8
#define SWEEP_TRANSPARENCY 0.6

/* texture parameters */
#define XTEX 256
#define YTEX 256
#define FIBRES 80
#define FIBRE_ACC 64
#define TEX_ROUGHNESS 0.1


/* Candle location */
GLfloat      points[ POINTS * 2 ];

/* mysterious variables ;) */
int view_angle = 55, yrot = 30, perspective = 20;
int graham_angle = 89;
int complete = 0, rounds = 0;

/* sweep list */
struct lister *head, *rot_pnt;

typedef struct lister {
    struct lister *next;
    GLfloat x, y;
    int angle;
} lister;


/* avoid namespace conflicts by defining everything static */
void vec_init();
static void fixlist();
static void generate_list();
static void init();
static void draw_table();
static void draw_candle();
static void generate_points();
static void check_nextpos();
static void graham_sweep();
static void graham_display( GLfloat upper, GLfloat lower );
static void graham_sweeper( GLfloat upper, GLfloat lower );
static void fibre_rec( float ptr[], int s, int e );
static void create_tex( GLubyte *ptr, int x, int y );


/* texturing */
GLubyte **tex;
GLuint texName;

/* list ids */
GLuint candle_list = 1, 
        table_list = 2;

/*
 * Main displaying function
 */
static void display()
{
    int i;

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glColor3f(1,1,1);      /* sane colors */

    /* Draw the table... */
    glCallList( table_list );

    /* Draw the candles... */
    for (i = 0; i < POINTS; i++) {
	glPushMatrix();
	glTranslatef( points[i*2], 0, points[i*2+1] ); /* center on candle */
	glCallList( candle_list );

	glPopMatrix();
    }
 
    /* Select transparent drawing, depth checks are disable because
     * the quads are not sorted... 
     */
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glEnable( GL_BLEND );
    glDisable( GL_CULL_FACE );
    glDisable( GL_LIGHTING );
    glDepthMask( GL_FALSE );

    graham_display( SWEEP_UPPER, SWEEP_LOWER );
    graham_sweeper( SWEEP_UPPER, SWEEP_LOWER );

    /* reset original values. */
    glDisable( GL_BLEND );
    glEnable( GL_CULL_FACE );
    glEnable( GL_LIGHTING );
    glDepthMask( GL_TRUE );
    glutSwapBuffers();
}


/* 
 * Do the complete part of the graham's by simply scanning the rotation
 * point list until currect point is encountered (or the whole list after 
 * the completion of the first sweep).
 */
static void graham_display( GLfloat upper, GLfloat lower )
{
    int i, a;
    lister *t;

    a = graham_angle - view_angle;
    if (a < 0)
	a += 360;

    glBegin(GL_QUAD_STRIP);
    t = head;
    i = complete+1;
    glColor4f( 0.5, 0.5, 1, 1 - pow(SWEEP_TRANSPARENCY,rounds+1) );
    while ( i-- || rounds ) {
	glVertex3f( t->x, lower, t->y );
	glVertex3f( t->x, upper, t->y );
	t = t->next;
	if (i <= 0)
	    glColor4f( 0.5, 0.5, 1, 1 - pow(SWEEP_TRANSPARENCY,rounds) );
	if (t == NULL)
	    break;
    }
    glEnd();
}


/* 
 * Do the graham's sweeper... Let the OpenGl do all the math, just rotate
 * and translate and draw "long" line towards x-inf...
 * stuupiid z-clipping or something,
 * makes it terminate before the vanishing point...
 */
static void graham_sweeper( GLfloat upper, GLfloat lower ) 
{
    int i;
    glColor4f( 0.5, 0.5, 1, 1 - SWEEP_TRANSPARENCY );
    glPushMatrix();
    glTranslatef( rot_pnt->x, 0, rot_pnt->y );
    glRotatef( graham_angle + 1, 0, 1, 0);      /* delayed sweeper, guards */
    glBegin(GL_QUAD_STRIP);                     /* againts roundoffs */
    for (i = 0; i < 200; i++) {
	glVertex3f( i, lower, 0 );
	glVertex3f( i, upper, 0 );
    }
    glEnd();
    glPopMatrix();
}


/*
 * Draws a single candle... (Used for list generation)
 * Glut has problem with lists.
 */
static void draw_candle()
{
    double a, b;
    int i;

    /* draw body */
    glColor3f( 1.0, 1.0, 1.0 );
    glBegin( GL_QUAD_STRIP );
    for (i = 0; i <= CANDLE_QUADS; i++) {
	a = CANDLE_RADIUS * sin( 2 * i * M_PI / CANDLE_QUADS );
	b = CANDLE_RADIUS * cos( 2 * i * M_PI / CANDLE_QUADS );
	glVertex3f( a, CANDLE_UPPER, b );
	glVertex3f( a, CANDLE_LOWER, b );
    }    
    glEnd();
    /* draw top */
    glColor3f( 0.8,0.8,0.8 );
    glBegin( GL_TRIANGLE_STRIP );
    for (i = 0; i <= CANDLE_QUADS; i++) {
	a = CANDLE_RADIUS * sin( 2 * i * M_PI / CANDLE_QUADS );
	b = CANDLE_RADIUS * cos( 2 * i * M_PI / CANDLE_QUADS );
	glVertex3f( 0, CANDLE_TOP, 0 );
	glVertex3f( a, CANDLE_UPPER, b);
    }
    glEnd();
    /* draw kernel */
    glColor3f( 0.1, 0.1, 0.1 );
    glBegin( GL_QUAD_STRIP );
    glVertex3f( KERNEL_RADIUS, CANDLE_TOP, 0 );
    glVertex3f( KERNEL_RADIUS, CANDLE_TOP+KERNEL_LEN, 0 );
    glVertex3f( 0, CANDLE_TOP, KERNEL_RADIUS );
    glVertex3f( 0, CANDLE_TOP+KERNEL_LEN, KERNEL_RADIUS );
    glVertex3f( -KERNEL_RADIUS, CANDLE_TOP, 0 );
    glVertex3f( -KERNEL_RADIUS, CANDLE_TOP+KERNEL_LEN, 0 );
    glVertex3f( 0, CANDLE_TOP, -KERNEL_RADIUS );
    glVertex3f( 0, CANDLE_TOP+KERNEL_LEN, -KERNEL_RADIUS );
    glEnd();
}


/*
 * Draws the table. The same texture is used twice... sometimes makes
 * stupid looking tables but creating two textures is just annoying... ;)
 * A bit inefficient looking code, too much done without arrays...
 * (used for list generation)
 */
static void draw_table() {
    int i;
    static GLfloat edges[8][3] = {
	{-2.0, 0, 1},  {-2.0, -0.1, 1}, {2.0, -0.1, 1}, {2.0, 0, 1},
	{2.0, 0, -1}, {2.0, -0.1, -1}, {-2.0, -0.1, -1}, {-2.0, 0, -1} };

    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
    glBindTexture(GL_TEXTURE_2D, texName);

    glBegin( GL_QUADS );
    /* Table... */
    glTexCoord2f(0.0, 0.0);  glVertex3f( 2.0,  0, -1 ); 
    glTexCoord2f(0.0, 1.0);  glVertex3f( -2.0, 0, -1 );
    glTexCoord2f(0.7, 1.0);  glVertex3f( -2.0, 0, 0 );
    glTexCoord2f(0.7, 0.0);  glVertex3f( 2.0, 0, 0 );
    glTexCoord2f(0.3, 0.0);  glVertex3f( 2.0,  0, 0 ); 
    glTexCoord2f(0.3, 1.0);  glVertex3f( -2.0, 0, 0 );
    glTexCoord2f(1.0, 1.0);  glVertex3f( -2.0, 0, 1 );
    glTexCoord2f(1.0, 0.0);  glVertex3f( 2.0, 0, 1 );

    /* Ends */
    glTexCoord2f(0.7, 0.0);  glVertex3f( -2.0,  0, -1 ); 
    glTexCoord2f(0.7, 0.1);  glVertex3f( -2.0, -0.1, -1 );
    glTexCoord2f(0.0, 0.1);  glVertex3f( -2.0, -0.1, 0 );
    glTexCoord2f(0.0, 0.0);  glVertex3f( -2.0, 0, 0 );
    glTexCoord2f(0.3, 0.0);  glVertex3f( -2.0,  0, 0 ); 
    glTexCoord2f(0.3, 0.1);  glVertex3f( -2.0, -0.1, 0 );
    glTexCoord2f(1.0, 0.1);  glVertex3f( -2.0, -0.1, 1 );
    glTexCoord2f(1.0, 0.0);  glVertex3f( -2.0, 0, 1 );

    glTexCoord2f(0.7, 0.0);  glVertex3f( 2.0,  -0.1, -1 ); 
    glTexCoord2f(0.7, 0.1);  glVertex3f( 2.0, 0, -1 );
    glTexCoord2f(0.0, 0.1);  glVertex3f( 2.0, 0, 0 );
    glTexCoord2f(0.0, 0.0);  glVertex3f( 2.0, -0.1, 0 );
    glTexCoord2f(0.3, 0.0);  glVertex3f( 2.0, -0.1, 0 ); 
    glTexCoord2f(0.3, 0.1);  glVertex3f( 2.0, 0, 0 );
    glTexCoord2f(1.0, 0.1);  glVertex3f( 2.0, 0, 1 );
    glTexCoord2f(1.0, 0.0);  glVertex3f( 2.0, -0.1, 1 );

    /* sides */
    glColor4ub( 200, 175, 29, 1.0 );
    for (i = 0; i < 8; i++)
	glVertex3fv( edges[i] );
    
    glEnd();
    glDisable(GL_TEXTURE_2D);
    
}


/*
 *
 */
static void keyboard( unsigned char key, int x, int y )
{
    switch (key) {
    case 27:
	exit(0);
	break;	
    }

}



/*
 * Window resize function
 */
static void reshape( int w, int h )
{
    GLfloat pos[] = {5,0,5,1};
    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glMatrixMode(GL_PROJECTION);


    glLoadIdentity();
    gluPerspective(60.0, (GLfloat) w/(GLfloat) h, 1.0, perspective );

    glLightfv( GL_LIGHT0, GL_POSITION, pos );
    glEnable( GL_LIGHT0 );
    glTranslatef (0.0, 0.0, -5.0);
    glRotatef( yrot, 1.0, 0.0, 0.0 );
    glRotatef( view_angle, 0.0, 1.0, 0.0 );
    glTranslatef (0.0, 0.0, 0.4 );

    glMatrixMode(GL_MODELVIEW);    
}



/* 
 * CALLME! Enter intro MainLoop after the call...
 */
void vec_init()
{
    init();
    generate_points();
    glutReshapeFunc(reshape);
    glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);
    glutIdleFunc( graham_sweep );

    /*
     * Prepare graham's...
     */
    generate_list();
    rot_pnt = head;
}


/*
 * initalization.
 */
static void init()
{
    glClearColor (0.0, 0.0, 0.0, 0.0);
    
    glShadeModel (GL_SMOOTH);
    
    glEnable( GL_CULL_FACE );
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    
    /* allocate texture space... */
    tex = (GLubyte **) malloc( 1 * sizeof(GLubyte *) );
    tex[0] = (GLubyte *) malloc( XTEX * YTEX * 3 * sizeof(GLubyte) );

    /* tex stuff */
    create_tex(tex[0], XTEX, YTEX);  
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glGenTextures(1, &texName);
    glBindTexture(GL_TEXTURE_2D, texName);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, XTEX, YTEX,
		 0, GL_RGB, GL_UNSIGNED_BYTE, tex[0]);

    glEnable( GL_DEPTH_TEST );

    /* create lists */
    glNewList( candle_list, GL_COMPILE ); draw_candle(); glEndList();
    glNewList( table_list, GL_COMPILE ); draw_table(); glEndList();

    
    glEnable( GL_LIGHTING );
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    glEnable( GL_COLOR_MATERIAL );
}



/***************************************************
 * Graham's back-end...
 **************************************************/

/*
 * Calculated the angle of two points.
 */
double calculate_angle( int a, int b )
{
    double yd, xd;       /* yd is calculated from z-coordinates */
    xd = points[ a * 2 ] - points[ b * 2 ];
    yd = points[ a * 2 + 1 ] - points[ b * 2 + 1 ];
    return atan2( yd, xd );
}



/*
 * Finds the starting point for Graham's sweep, (min x, if two equal, max z)
 */
int find_start_point()
{
    int i, p;
    p = 0;
    for (i = 1; i < POINTS; i++) {
	if (points[ i * 2 ] < points[ p * 2 ] ||
	    (points[ i * 2 ] == points[ p * 2 ] && 
	     points[ i * 2 + 1 ] > points[ p * 2 + 1 ]))
	    p = i;
    }
    assert( p != -1 );

    return p;
}


/*
 * Finds the next point for graham's algorithm.
 */
int find_next_point( int point, double last_angle )
{
    double t, min;
    int i, p;
    p = -1;

    for (i = 0; i < POINTS; i++) {
	if (i != point ) {
	    t = calculate_angle( point, i );
	    while (t < last_angle)
		t += 2*M_PI;
	    
	    if (p == -1 || t < min) {
		p = i;
		min = t;
	    }
	}
    }

    assert (p != -1);
    return p;
}


/*
 * Sweep node creation.
 */
static void next_node( lister **ptr, GLfloat x, GLfloat y )
{
    *ptr = (lister *) malloc( sizeof(lister) );
    (*ptr)->x = x;
    (*ptr)->y = y;
}


/*
 * Generates Graham's sweep blocks... From point-to-point, the candle
 * "avoidance" is done later.
 */
static void generate_list()
{
    int s, p, n;
    double last_angle;
    lister **ptr;

    p = s = find_start_point();
    ptr = &head;
    next_node( ptr, points[ s * 2 ], points[ s * 2 + 1 ] );
    last_angle = graham_angle;

    while (1) {
	n = find_next_point( p, last_angle );
	last_angle = calculate_angle( p, n );

	(*ptr)->angle = 180 - (int)(last_angle * 180.0 / M_PI);
	while ((*ptr)->angle < 0)
	    (*ptr)->angle += 360;

	ptr = &((*ptr)->next);
	next_node( ptr, points[ n * 2 ], points[ n * 2 + 1 ] );

	if (n == s) {
	    break;
	} else {
	    p = n;
	}
    }
    /* Create copy of the first for easier looping... */
    next_node( ptr, head->x, head->y );
    (*ptr)->angle = graham_angle;        /* termination angle */
    (*ptr)->next = NULL;

    fixlist();
}


/*
 * Now we must scan through the list and move the coordinates outside
 * the candles... This splits sweep area into small blocks when going around 
 * candles...
 */
static void fixlist()
{
    int terminate, last_ang, overflows, old_ang;
    double a, b;
    lister *t, *temp;

    if (head->angle == 90)                   /* BUG FIX */
	graham_angle = 90;

    last_ang = graham_angle;


    for (t = head; t != NULL; t = t->next) {
	
	printf("next point ang: %i/%i \t(%f %f)\n", last_ang, t->angle,
	  t->x, t->y );
	
	overflows =  (last_ang < t->angle);  /* true when wraparound */
	terminate = (last_ang == t->angle);  /* true when tangent */
	while ( 1 ) {
	    a = SWEEP_RADIUS * sin( last_ang * M_PI / 180.0 );
	    b = SWEEP_RADIUS * cos( last_ang * M_PI / 180.0 );
	    if (terminate)
		break;

	    old_ang = last_ang;
	    last_ang -= 360 / (CANDLE_QUADS * 2);
	    while (last_ang < 0) {
		if (!overflows) {            /* overflow when not "possible" */
		    terminate = 1;           /* indicates block termination */
		    old_ang = last_ang = t->angle;
		    break;
		}
		overflows = 0;               /* set overflow 'passed' */
		last_ang += 360;
	    }

	    
	    /*printf("angle: %i %i\n", last_ang, t->angle );*/
	    
	    if (last_ang < t->angle && !overflows) {
		old_ang = last_ang = t->angle;		
		terminate = 1;
	    }

	    next_node( &temp, t->x, t->y );
	    t->x -= a; t->y -= b;
	    temp->angle = t->angle;
	    t->angle = old_ang;
	    
	    temp->next = t->next;
	    t->next = temp;
	    t = t->next;
	}
	t->x -= a; t->y -= b;
    }
}


/*
 * Generate the "candles" as points...
 */
static void generate_points( void )
{
    int i;
    for (i = 0; i < POINTS; i++) {
	points[i * 2 + 0] = ((AREA_MAXX - AREA_MINX) * rand()) / 
	    (RAND_MAX +1.0) + AREA_MINX;
	points[i * 2 + 1] = ((AREA_MAXZ - AREA_MINZ) * rand()) / 
	    (RAND_MAX +1.0) + AREA_MINZ;
    }
}


/*
 * Checks if the next "block" is complete...
 */
static void check_nextpos()
{
    while (graham_angle == rot_pnt->angle) {
	rot_pnt = rot_pnt->next;
	complete++;
	if (rot_pnt == NULL) {
	    rot_pnt = head;
	    rounds++;
	    complete = 0;
	}
    }
}


/*
 * Advances the sweep...
 */
static void graham_sweep()
{
    check_nextpos();

    graham_angle -= 1;
    while (graham_angle < 0)
	graham_angle += 360;

    /*    check_nextpos();*/

    printf( "angle: %i/%i\n", graham_angle,  rot_pnt->angle );
    glutPostRedisplay();
}


/************************************
 * Texture generation
 ************************************/

/*
 * Random midpoint algo... (r[s]+r[e])/2 * |e - s| * roughness * r
 *
 * r = random (-0.5...0.5)
 */
static void fibre_rec( float ptr[], int s, int e )
{
    float r = TEX_ROUGHNESS, t;
    int m;
    m = (e + s) / 2;
    if (s == e || s == m)
	return;                  /* base case */
    t = (ptr[e]+ptr[s])/2 + r * (e-s) * (((float)rand())/RAND_MAX-0.5);
    ptr[m] = (t < 0 ? 0 : t);

    fibre_rec( ptr, s, m );
    fibre_rec( ptr, m, e );
}


/*
 * Creates table textures. First the wood fibres are generated using random
 * midpoint algorithm (recursive implementation).
 */
static void create_tex( GLubyte *ptr, int x, int y )
{
    int i, j, c, t;
    float fibre[ FIBRES ][ FIBRE_ACC ];

    /* generate fibres */
    for (c = 0; c < FIBRES; c++) {
	fibre[c][0] = fibre[c][ FIBRE_ACC - 1 ] = x*c / FIBRES;
	fibre_rec(fibre[c], 0, FIBRE_ACC - 1 );
    }

    /* generate texture */
    for (i = 0; i < y; i++) {
	for (j = 0; j < x; j++) {
	    ptr[ 3*(i * x + j) + 0 ] = 190;
	    ptr[ 3*(i * x + j) + 1 ] = 165;
	    ptr[ 3*(i * x + j) + 2 ] = 29;
	}
    }
    /* draw fibres into tex */
    for (c = 0; c < FIBRES; c++) {
	for (i = 0; i < y; i++) {
	    j = fibre[ c ][ i * FIBRE_ACC / y ];
	    ptr[ 3*(i * x + j) + 0 ] = 150;
	    ptr[ 3*(i * x + j) + 1 ] = 125;
	    ptr[ 3*(i * x + j) + 2 ] = 20;
	}
    }

    /* smooth. */
    for (i = 0; i < y-1; i++)
	for (j = 0; j < x-1; j++) {
	    for (c = 0; c < 3; c++) {
		t = (ptr[ 3*(i * x + j) + c ] +
		     ptr[ 3*((i+1) * x + j) + c ] +
		     ptr[ 3*(i * x + j+1) + c ] +
		     ptr[ 3*((i+1) * x + j) + c ]) / 4;
		assert ( 0 <= t && t <= 255 );
		ptr[ 3*(i * x + j) + c ] = t;
	    }
	}
}


/*
 * Main Loop 
 */
int main(int argc, char **argv)
{
    srand((unsigned int)time( NULL ));

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    glutInitWindowSize(640, 480);
    glutInitWindowPosition(0,0);
    glutCreateWindow(argv[0]);

    vec_init();
    glutMainLoop(); 
    return 0;
}

